from typing import Union, Literal, Callable, Any
from warnings import warn

import shapely
import pandas as pd
import geopandas as gpd

import rhealpixdggs.rhp_wrappers as rhp_py

from .util.const import *
from .util.functools import wrapped_partial

AnyDataFrame = Union[pd.DataFrame, gpd.GeoDataFrame]


@pd.api.extensions.register_dataframe_accessor("rhp")
class rHPAccessor:
    """
    Shamelessly appropriated from equivalent class in h3pandas package

    The h3pandas repo is found here: https://github.com/DahnJ/H3-Pandas

    TODO: - Support both plane and sphere (currently hardwired to sphere)
          - Support user defined instance of underlying dggs (currently hardwired to WGS84_003)
    """

    def __init__(self, df: pd.DataFrame) -> None:
        self._df = df

    # RHP wrapper API
    def geo_to_rhp(
        self,
        resolution: int,
        lat_col: str = "lat",
        lng_col: str = "lng",
        set_index: bool = True,
    ) -> AnyDataFrame:
        """
        Adds rHEALPix index to (Geo)DataFrame

        pd.DataFrame: uses `lat_col` and `lng_col` (default `lat` and `lng`)
        gpd.GeoDataFrame: uses `geometry`

        resolution : int
            rHEALPix resolution
        lat_col : str
            Name of the latitude column (if used), default 'lat'
        lng_col : str
            Name of the longitude column (if used), default 'lng'
        set_index : bool
            If True, the columns with rHEALPix addresses is set as index, default 'True'

        Returns
        -------
        (Geo)DataFrame with rHEALPix addresses added
        """

        # DataFrame wrangling
        if isinstance(self._df, gpd.GeoDataFrame):
            lngs = self._df.geometry.x
            lats = self._df.geometry.y
        else:
            lngs = self._df[lng_col]
            lats = self._df[lat_col]

        # Index conversion
        rhpaddresses = [
            rhp_py.geo_to_rhp(lat, lng, resolution, False)
            for lat, lng in zip(lats, lngs)
        ]

        # Add results to DataFrame
        colname = f"{COLUMNS['prefix']}{resolution:02}"
        assign_arg = {colname: rhpaddresses}
        df = self._df.assign(**assign_arg)
        if set_index:
            return df.set_index(colname)
        return df

    def rhp_to_geo(self) -> gpd.GeoDataFrame:
        """Add `geometry` with centroid of each rHEALPix address to the DataFrame.
        Assumes rHEALPix index.

        Returns
        -------
        GeoDataFrame with Point geometry

        See Also
        --------
        rhp_to_geo_boundary : Adds an rHEALPix cell
        """
        return self._apply_index_assign(
            wrapped_partial(rhp_py.rhp_to_geo, geo_json=True, plane=False),
            COLUMNS["geometry"],
            lambda x: shapely.geometry.Point(x),
            lambda x: gpd.GeoDataFrame(
                x, crs="epsg:4326"
            ),  # TODO: add correct coordinate system?
        )

    def rhp_to_geo_boundary(self) -> AnyDataFrame:
        """Add `geometry` with rHEALPix squares to the DataFrame. Assumes rHEALPix index.

        Returns
        -------
        GeoDataFrame with rHEALPix geometry
        """
        return self._apply_index_assign(
            wrapped_partial(rhp_py.rhp_to_geo_boundary, geo_json=True, plane=False),
            COLUMNS["geometry"],
            lambda x: shapely.geometry.Polygon(x),
            lambda x: gpd.GeoDataFrame(
                x, crs="epsg:4326"
            ),  # TODO: add correct coordinate system?
        )

    def rhp_get_resolution(self) -> AnyDataFrame:
        """
        Adds a column 'rhp_resolution' with the resolution of each cell to the dataframe.
        """
        return self._apply_index_assign(
            rhp_py.rhp_get_resolution, COLUMNS["resolution"]
        )

    def rhp_get_base_cell(self) -> AnyDataFrame:
        """
        Adds a column 'rhp_base_cell' with the resolution 0 parent cell to the dataframe.
        """
        return self._apply_index_assign(rhp_py.rhp_get_base_cell, COLUMNS["base_cell"])

    def rhp_is_valid(self) -> AnyDataFrame:
        """
        Adds a column 'rhp_is_valid' indicating if the cell addresses are valid rHEALPix
        addresses or not.
        """
        return self._apply_index_assign(rhp_py.rhp_is_valid, COLUMNS["is_valid"])

    def k_ring(
        self, k: int = 1, explode: bool = False, verbose: bool = True
    ) -> AnyDataFrame:
        """
        Parameters
        ----------
        k : int
            the distance from the origin rHEALPix address. Default k = 1
        explode : bool
            If True, will explode the resulting list vertically.
            All other columns' values are copied.
            Default: False

        TODO: find out if rhp needs the following note (and the referenced function)
        See Also
        --------
        k_ring_smoothing : Extended API method that distributes numeric values
            to the k-ring cells
        """
        if verbose:
            warn(str.format(rhp_py.CELL_RING_WARNING, "k"))

        func = wrapped_partial(rhp_py.k_ring, k=k, verbose=False)
        column_name = COLUMNS["k_ring"]
        if explode:
            return self._apply_index_explode(func, column_name, list)
        return self._apply_index_assign(func, column_name, list)

    def cell_ring(
        self, k: int = 1, explode: bool = False, verbose: bool = True
    ) -> AnyDataFrame:
        """
        Adds a column 'rhp_cell_ring' of cells at distance k from the existing entries
        to the dataframe.

        explode = False will add the cell ring as a list associated with the existing
        entry.

        explode = True will add the cell ring one cell at a time (repeating existing
        entries).
        """
        if verbose:
            warn(str.format(rhp_py.CELL_RING_WARNING, "cell"))

        func = wrapped_partial(rhp_py.cell_ring, k=k, verbose=False)
        column_name = COLUMNS["cell_ring"]
        if explode:
            return self._apply_index_explode(func, column_name, list)
        return self._apply_index_assign(func, column_name, list)

    def rhp_to_parent(self, resolution: int = None) -> AnyDataFrame:
        """
        Adds a column 'rhp_parent' with the parent cell at the requested resolution to the
        dataframe.
        ----------
        Parameters
        ----------
        resolution : int or None
            rHEALPix resolution. If None, then returns the direct parent of each rHEALPix cell.
        """
        column = (
            f"{COLUMNS['prefix']}{resolution:02}"
            if resolution is not None
            else COLUMNS["parent"]
        )

        return self._apply_index_assign(
            wrapped_partial(rhp_py.rhp_to_parent, res=resolution), column
        )

    def rhp_to_center_child(self, resolution: int = None) -> AnyDataFrame:
        """
        Adds a column 'rhp_center_child' with the address of the central child cell at the
        requested resolution to the dataframe.
        ----------
        Parameters
        ----------
        resolution : int or None
            rHEALPix resolution. If none, then returns the child of resolution directly
            below that of each rHEALPix cell
        """
        return self._apply_index_assign(
            wrapped_partial(rhp_py.rhp_to_center_child, res=resolution),
            COLUMNS["center_child"],
        )

    def polyfill(
        self,
        resolution: int,
        explode: bool = False,
        compress: bool = False,
        verbose: bool = True,
    ) -> AnyDataFrame:
        """
        Adds a column 'rhp_polyfill' listing the cells where the cell centroid is
        contained within the input (multi)polygon geometry.
        Relies on the dataframe having a 'geometry' column with shapely Polygon or
        MultiPolygon entries.
        ----------
        Parameters
        ----------
        resolution : int
            rHEALPix resolution
        explode : bool
            If True, will explode the resulting list vertically.
            All other columns' values are copied.
            Default: False
        compress : bool
            If True, will group cells making up a parent cell and use the address of the
            parent cell instead of the individual child addresses.
            Default: False
        """

        # Need to wrap polyfill for each dataframe row so we can call it with geometry
        def func(row):
            if not COLUMNS["geometry"] in row.keys():
                return None
            else:
                return rhp_py.polyfill(
                    row[COLUMNS["geometry"]], resolution, False, compress, verbose
                )

        # Polyfill cell sets for each row
        result = self._df.apply(func, axis=1)

        if not explode:
            assign_args = {COLUMNS["polyfill"]: result}
            return self._df.assign(**assign_args)

        result = result.explode().to_frame(COLUMNS["polyfill"])

        return self._df.join(result)

    def cell_area(self, unit: Literal["km^2", "m^2"] = "km^2") -> AnyDataFrame:
        """
        Adds a column 'rhp_cell_area' to the dataframe of cells addresses.
        ----------
        Parameters
        ----------
        unit : str, options: 'km^2' or 'm^2'
            Unit for area result. Default: 'km^2`

        TODO: find out the meaning of unit "rads^2" that appears in h3pandas
        """
        return self._apply_index_assign(
            wrapped_partial(rhp_py.cell_area, unit=unit), COLUMNS["cell_area"]
        )

    def linetrace(self) -> AnyDataFrame:
        raise NotImplementedError()

    # Aggregate functions
    def geo_to_rhp_aggregate(
        self,
        resolution: int,
        operation: Union[dict, str, Callable] = "sum",
        lat_col: str = "lat",
        lng_col: str = "lng",
        return_geometry: bool = True,
    ) -> pd.DataFrame:
        """Adds rHEALPix index to DataFrame, groups points with the same index
        and performs `operation`.

        pd.DataFrame: uses `lat_col` and `lng_col` (default `lat` and `lng`)
        gpd.GeoDataFrame: uses `geometry`

        Parameters
        ----------
        resolution : int
            rHEALPix resolution
        operation : Union[dict, str, Callable]
            Argument passed to DataFrame's `agg` method, default 'sum'
        lat_col : str
            Name of the latitude column (if used), default 'lat'
        lng_col : str
            Name of the longitude column (if used), default 'lng'
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the square cells.
            Default = True

        Returns
        -------
        (Geo)DataFrame aggregated by rHEALPix address into which each row's point falls

        See Also
        --------
        geo_to_rhp : rHEALPix API method upon which this function builds

        """
        colname = (
            f"{COLUMNS['prefix']}{resolution:02}"
            if resolution is not None
            else COLUMNS["parent"]
        )
        grouped = pd.DataFrame(
            self.geo_to_rhp(resolution, lat_col, lng_col, False)
            .drop(columns=[lat_col, lng_col, COLUMNS["geometry"]], errors="ignore")
            .groupby(colname)
            .agg(operation)
        )

        return grouped.rhp.rhp_to_geo_boundary() if return_geometry else grouped

    def rhp_to_parent_aggregate(
        self,
        resolution: int,
        operation: Union[dict, str, Callable] = "sum",
        return_geometry: bool = True,
    ) -> gpd.GeoDataFrame:
        """Assigns parent cell to each row, groups by it and performs `operation`.
        Assumes rHEALPix index.

        Parameters
        ----------
        resolution : int
            rHEALPix resolution
        operation : Union[dict, str, Callable]
            Argument passed to DataFrame's `agg` method, default 'sum'
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the square cells.
            Default = True

        Returns
        -------
        (Geo)DataFrame aggregated by the parent of each rHEALPix address

        See Also
        --------
        rhp_to_parent : rHEALPix API method upon which this function builds

        """
        parent_rhpaddresses = [
            rhp_py.rhp_to_parent(rhpaddress, resolution)
            for rhpaddress in self._df.index
        ]
        rhp_parent_column = (
            f"{COLUMNS['prefix']}{resolution:02}"
            if resolution is not None
            else COLUMNS["parent"]
        )
        kwargs_assign = {rhp_parent_column: parent_rhpaddresses}
        grouped = (
            self._df.assign(**kwargs_assign)
            .groupby(rhp_parent_column)[
                [c for c in self._df.columns if c != COLUMNS["geometry"]]
            ]
            .agg(operation)
        )

        return grouped.rhp.rhp_to_geo_boundary() if return_geometry else grouped

    def polyfill_resample(
        self,
        resolution: int,
        return_geometry: bool = True,
        compress: bool = False,
        verbose: bool = True,
    ) -> AnyDataFrame:
        """Experimental as stated in h3pandas, where this function comes from.
        Currently essentially polyfill(..., explode=True) that sets the rHEALPix
        index and adds the rHEALPix cell geometry if requested by return_geometry.

        Parameters
        ----------
        resolution : int
            rHEALPix resolution
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the hexagonal cells.
            Default = True
        compress : bool
            (Optional) Whether to compress cell groups that make up a parent cell
            into the parent, returning 1 id instead of 9.
            Default = False

        Returns
        -------
        (Geo)DataFrame with rHEALPix cells with centroids within the input polygons.

        See Also
        --------
        polyfill : rHEALPix API method upon which this method builds
        """
        result = self._df.rhp.polyfill(
            resolution, explode=True, compress=compress, verbose=verbose
        )
        uncovered_rows = result[COLUMNS["polyfill"]].isna()
        n_uncovered_rows = uncovered_rows.sum()
        if verbose and (n_uncovered_rows > 0):
            warn(
                f"{n_uncovered_rows} rows did not generate a cell."
                "Consider using a finer resolution."
            )
            result = result.loc[~uncovered_rows]

        result = result.reset_index().set_index(COLUMNS["polyfill"])

        return result.rhp.rhp_to_geo_boundary() if return_geometry else result

    # Helper functions
    def _apply_index_assign(
        self,
        func: Callable,
        column_name: str,
        processor: Callable = lambda x: x,
        finalizer: Callable = lambda x: x,
    ) -> Any:
        """
        Helper method. Applies `func` to index and assigns the result to `column`.

        Parameters
        ----------
        func : Callable
            single-argument function to be applied to each rHEALPix address
        column_name : str
            name of the resulting column
        processor : Callable
            (Optional) further processes the result of func. Default: identity
        finalizer : Callable
            (Optional) further processes the resulting dataframe. Default: identity

        Returns
        -------
        Dataframe with column `column` containing the result of `func`.
        If using `finalizer`, can return anything the `finalizer` returns.
        """
        result = [processor(func(rhpaddress)) for rhpaddress in self._df.index]
        assign_args = {column_name: result}

        return finalizer(self._df.assign(**assign_args))

    def _apply_index_explode(
        self,
        func: Callable,
        column_name: str,
        processor: Callable = lambda x: x,
        finalizer: Callable = lambda x: x,
    ) -> Any:
        """Helper method. Applies a list-making `func` to index and performs
        a vertical explode.
        Any additional values are simply copied to all the rows.

        Parameters
        ----------
        func : Callable
            single-argument function to be applied to each H3 address
        column_name : str
            name of the resulting column
        processor : Callable
            (Optional) further processes the result of func. Default: identity
        finalizer : Callable
            (Optional) further processes the resulting dataframe. Default: identity

        Returns
        -------
        Dataframe with column `column` containing the result of `func`.
        If using `finalizer`, can return anything the `finalizer` returns.
        """
        result = (
            pd.DataFrame.from_dict(
                {
                    rhpaddress: processor(func(rhpaddress))
                    for rhpaddress in self._df.index
                },
                orient="index",
            )
            .stack()
            .to_frame(column_name)
            .reset_index(level=1, drop=True)
        )
        result = self._df.join(result)
        return finalizer(result)

    def _multiply_numeric(self):
        raise NotImplementedError()



from functools import partial, update_wrapper


def wrapped_partial(func, *args, **kwargs):
    """
    Properly wrapped partial function

    Appropriated from h3pandas.util package
    """
    partial_func = partial(func, *args, **kwargs)
    update_wrapper(partial_func, func)
    return partial_func



COLUMNS = {
    "prefix": "rhp_",
    "is_valid": "rhp_is_valid",
    "resolution": "rhp_resolution",
    "base_cell": "rhp_base_cell",
    "parent": "rhp_parent",
    "center_child": "rhp_center_child",
    "cell_area": "rhp_cell_area",
    "cell_ring": "rhp_cell_ring",
    "k_ring": "rhp_k_ring",
    "polyfill": "rhp_polyfill",
    "linetrace": "rhp_linetrace",
    "geometry": "geometry",
}



import pytest

import pandas as pd
import geopandas as gpd

from geopandas.testing import assert_geodataframe_equal
from shapely.geometry import box, Polygon

from rhppandas import rhppandas
from rhppandas.util.const import *
from rhealpixdggs import rhp_wrappers as rhp_py

pd.option_context("display.float_format", "{:0.15f}".format)
pd.set_option("display.precision", 15)


# Fixtures
@pytest.fixture
def basic_dataframe():
    """DataFrame with lat and lng columns"""
    return pd.DataFrame({"lat": [50, 51], "lng": [14, 15]})


@pytest.fixture
def basic_dataframe_with_values(basic_dataframe):
    """DataFrame with lat and lng columns and values"""
    return basic_dataframe.assign(val=[2, 5])


@pytest.fixture
def basic_geodataframe(basic_dataframe):
    """GeoDataFrame with POINT geometry"""
    geometry = gpd.points_from_xy(basic_dataframe["lng"], basic_dataframe["lat"])
    return gpd.GeoDataFrame(geometry=geometry, crs="epsg:4326")


@pytest.fixture
def basic_geodataframe_with_values(basic_geodataframe):
    """GeoDataFrame with POINT geometry and values"""
    return basic_geodataframe.assign(val=[2, 5])


@pytest.fixture
def basic_geodataframe_polygon():
    """GeoDataFrame with POLYGON geometry"""
    geom = box(0, 0, 1, 1)
    return gpd.GeoDataFrame(geometry=[geom], crs="epsg:4326")


@pytest.fixture
def basic_geodataframe_polygons():
    """GeoDataFrame with POLYGON geometries"""
    geoms = [box(14, 50, 15, 51), box(14, 50, 15, 52)]
    return gpd.GeoDataFrame(geometry=geoms, crs="epsg:4326")


@pytest.fixture
def indexed_dataframe(basic_dataframe):
    """DataFrame with lat, lng and resolution 9 rHEALPix index"""
    return basic_dataframe.assign(rhp_09=["N216055611", "N208542111"]).set_index(
        f"{COLUMNS['prefix']}09"
    )


@pytest.fixture
def indexed_dataframe_centre_cells(basic_dataframe):
    """DataFrame with lat, lng and resolution 10 rHEALPix index"""
    return basic_dataframe.assign(rhp_10=["N2160556114", "N2085421114"]).set_index(
        f"{COLUMNS['prefix']}10"
    )


@pytest.fixture
def rhp_dataframe_with_values():
    """DataFrame with resolution 9 rHEALPix index and values"""
    index = ["N216055611", "N216055612", "N216055615"]
    return pd.DataFrame({"val": [1, 2, 5]}, index=index)


@pytest.fixture
def rhp_geodataframe_with_values(rhp_dataframe_with_values):
    """GeoDataFrame with resolution 9 rHEALPix index, values, and cell geometries"""
    geometry = [
        Polygon(rhp_py.rhp_to_geo_boundary(h, True, False))
        for h in rhp_dataframe_with_values.index
    ]
    return gpd.GeoDataFrame(
        rhp_dataframe_with_values, geometry=geometry, crs="epsg:4326"
    )


# Tests: rHEALPix wrapper API
class TestGeoToRhp:
    def test_geo_to_rhp(self, basic_dataframe):
        result = basic_dataframe.rhp.geo_to_rhp(9)
        expected = basic_dataframe.assign(
            rhp_09=["N216055147", "N208518546"]
        ).set_index(f"{COLUMNS['prefix']}09")

        pd.testing.assert_frame_equal(expected, result)

    def test_geo_to_rhp_geo(self, basic_geodataframe):
        result = basic_geodataframe.rhp.geo_to_rhp(9)
        expected = basic_geodataframe.assign(
            rhp_09=["N216055147", "N208518546"]
        ).set_index(f"{COLUMNS['prefix']}09")

        pd.testing.assert_frame_equal(expected, result)

    def test_geo_to_rhp_polygon(self, basic_geodataframe_polygon):
        with pytest.raises(ValueError):
            basic_geodataframe_polygon.rhp.geo_to_rhp(9)


class TestRhpToGeo:
    def test_rhp_to_geo(self, indexed_dataframe):
        lats = [50.06543285982062, 51.06479381112059]
        lngs = [14.000847727642311, 14.998138688394175]
        geometry = gpd.points_from_xy(x=lngs, y=lats, crs="epsg:4326")
        expected = gpd.GeoDataFrame(indexed_dataframe, geometry=geometry)
        result = indexed_dataframe.rhp.rhp_to_geo()

        assert_geodataframe_equal(expected, result, check_less_precise=True)


class TestRhpToGeoBoundary:
    def test_rhp_to_geo_boundary(self, indexed_dataframe):
        c1 = (
            (13.996245382425958, 50.067960668809754),
            (14.0016956337431, 50.067960668809754),
            (14.00544959128063, 50.062905036612165),
            (13.999999999999975, 50.062905036612165),
            (13.996245382425958, 50.067960668809754),
        )
        c2 = (
            (14.993485139914386, 51.06731331121636),
            (14.999069305702065, 51.06731331121636),
            (15.002791736460077, 51.06227429726551),
            (14.997208263539921, 51.06227429726551),
            (14.993485139914386, 51.06731331121636),
        )
        geometry = [Polygon(c1), Polygon(c2)]

        result = indexed_dataframe.rhp.rhp_to_geo_boundary()
        expected = gpd.GeoDataFrame(
            indexed_dataframe, geometry=geometry, crs="epsg:4326"
        )

        assert_geodataframe_equal(expected, result, check_less_precise=True)

    def test_rhp_to_geo_boundary_wrong_index(self, indexed_dataframe):
        c = (
            (13.996245382425958, 50.067960668809754),
            (14.0016956337431, 50.067960668809754),
            (14.00544959128063, 50.062905036612165),
            (13.999999999999975, 50.062905036612165),
            (13.996245382425958, 50.067960668809754),
        )
        geometry = [Polygon(c), Polygon()]
        indexed_dataframe.index = [str(indexed_dataframe.index[0])] + ["invalid"]
        result = indexed_dataframe.rhp.rhp_to_geo_boundary()
        expected = gpd.GeoDataFrame(
            indexed_dataframe, geometry=geometry, crs="epsg:4326"
        )

        assert_geodataframe_equal(expected, result, check_less_precise=True)


class TestRhpGetResolution:
    def test_rhp_get_resolution(self, rhp_dataframe_with_values):
        expected = rhp_dataframe_with_values.assign(rhp_resolution=9)
        result = rhp_dataframe_with_values.rhp.rhp_get_resolution()

        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_get_resolution_index_only(self, rhp_dataframe_with_values):
        del rhp_dataframe_with_values["val"]
        expected = rhp_dataframe_with_values.assign(rhp_resolution=9)
        result = rhp_dataframe_with_values.rhp.rhp_get_resolution()

        pd.testing.assert_frame_equal(expected, result)


class TestRhpGetBaseCell:
    def test_rhp_get_base_cell(self, indexed_dataframe):
        expected = indexed_dataframe.assign(rhp_base_cell=["N", "N"])
        result = indexed_dataframe.rhp.rhp_get_base_cell()

        pd.testing.assert_frame_equal(expected, result)


class TestRhpIsValid:
    def test_rhp_is_valid(self, indexed_dataframe):
        indexed_dataframe.index = [str(indexed_dataframe.index[0])] + ["invalid"]
        expected = indexed_dataframe.assign(rhp_is_valid=[True, False])
        result = indexed_dataframe.rhp.rhp_is_valid()

        pd.testing.assert_frame_equal(expected, result)


class TestKRing:
    def test_rhp_0_ring(self, indexed_dataframe_centre_cells):
        expected = indexed_dataframe_centre_cells.assign(
            rhp_k_ring=[[h] for h in indexed_dataframe_centre_cells.index]
        )
        result = indexed_dataframe_centre_cells.rhp.k_ring(0, verbose=False)
        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_k_ring(self, indexed_dataframe_centre_cells):
        expected_indices = [
            {
                "N2160556114",
                "N2160556110",
                "N2160556111",
                "N2160556112",
                "N2160556115",
                "N2160556118",
                "N2160556117",
                "N2160556116",
                "N2160556113",
            },
            {
                "N2085421114",
                "N2085421110",
                "N2085421111",
                "N2085421112",
                "N2085421115",
                "N2085421118",
                "N2085421117",
                "N2085421116",
                "N2085421113",
            },
        ]
        expected = indexed_dataframe_centre_cells.assign(rhp_k_ring=expected_indices)
        result = indexed_dataframe_centre_cells.rhp.k_ring(verbose=False)
        result[COLUMNS["k_ring"]] = result[COLUMNS["k_ring"]].apply(
            set
        )  # Convert to set for testing
        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_k_ring_explode(self, indexed_dataframe_centre_cells):
        expected_indices = set().union(
            *[
                {
                    "N2160556114",
                    "N2160556110",
                    "N2160556111",
                    "N2160556112",
                    "N2160556115",
                    "N2160556118",
                    "N2160556117",
                    "N2160556116",
                    "N2160556113",
                },
                {
                    "N2085421114",
                    "N2085421110",
                    "N2085421111",
                    "N2085421112",
                    "N2085421115",
                    "N2085421118",
                    "N2085421117",
                    "N2085421116",
                    "N2085421113",
                },
            ]
        )
        result = indexed_dataframe_centre_cells.rhp.k_ring(explode=True, verbose=False)
        assert len(result) == len(indexed_dataframe_centre_cells) * 9
        assert set(result[COLUMNS["k_ring"]]) == expected_indices
        assert not result["lat"].isna().any()


class TestCellRing:
    pass


class TestRhpToParent:
    def test_rhp_to_parent_level_1(self, rhp_dataframe_with_values):
        rhp_parent = "N2"
        result = rhp_dataframe_with_values.rhp.rhp_to_parent(1)
        expected = rhp_dataframe_with_values.assign(rhp_01=rhp_parent)

        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_to_direct_parent(self, rhp_dataframe_with_values):
        rhp_parents = ["N21605561", "N21605561", "N21605561"]
        result = rhp_dataframe_with_values.rhp.rhp_to_parent()
        expected = rhp_dataframe_with_values.assign(rhp_parent=rhp_parents)

        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_to_parent_level_0(self, rhp_dataframe_with_values):
        rhp_parent = "N"
        result = rhp_dataframe_with_values.rhp.rhp_to_parent(0)
        expected = rhp_dataframe_with_values.assign(rhp_00=rhp_parent)

        pd.testing.assert_frame_equal(expected, result)


class TestRhpToCenterChild:
    def test_rhp_to_center_child(self, indexed_dataframe):
        expected = indexed_dataframe.assign(
            rhp_center_child=["N216055611444", "N208542111444"]
        )
        result = indexed_dataframe.rhp.rhp_to_center_child(13)

        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_to_center_child_one_level(self, indexed_dataframe):
        expected = indexed_dataframe.assign(
            rhp_center_child=["N2160556114", "N2085421114"]
        )
        result = indexed_dataframe.rhp.rhp_to_center_child()

        pd.testing.assert_frame_equal(expected, result)

    def test_rhp_to_center_child_wrong_index(self, indexed_dataframe):
        indexed_dataframe.index = [str(indexed_dataframe.index[0])] + ["invalid"]
        result = indexed_dataframe.rhp.rhp_to_center_child()
        expected = indexed_dataframe.assign(rhp_center_child=["N2160556114", None])

        pd.testing.assert_frame_equal(expected, result)


class TestPolyfill:
    def test_empty_polyfill(self, rhp_geodataframe_with_values):
        expected = rhp_geodataframe_with_values.assign(
            rhp_polyfill=[set(), set(), set()]
        )
        result = rhp_geodataframe_with_values.rhp.polyfill(1)
        assert_geodataframe_equal(expected, result)

    def test_polyfill(self, rhp_geodataframe_with_values):
        expected_cells = [
            {
                "N2160556110",
                "N2160556111",
                "N2160556112",
                "N2160556113",
                "N2160556114",
                "N2160556115",
                "N2160556116",
                "N2160556117",
                "N2160556118",
            },
            {
                "N2160556120",
                "N2160556121",
                "N2160556122",
                "N2160556123",
                "N2160556124",
                "N2160556125",
                "N2160556126",
                "N2160556127",
                "N2160556128",
            },
            {
                "N2160556150",
                "N2160556151",
                "N2160556152",
                "N2160556153",
                "N2160556154",
                "N2160556155",
                "N2160556156",
                "N2160556157",
                "N2160556158",
            },
        ]
        expected = rhp_geodataframe_with_values.assign(rhp_polyfill=expected_cells)
        result = rhp_geodataframe_with_values.rhp.polyfill(10)
        assert_geodataframe_equal(expected, result)

    def test_polyfill_explode(self, rhp_geodataframe_with_values):
        expected_indices = set().union(
            *[
                [
                    "N2160556110",
                    "N2160556111",
                    "N2160556112",
                    "N2160556113",
                    "N2160556114",
                    "N2160556115",
                    "N2160556116",
                    "N2160556117",
                    "N2160556118",
                ],
                [
                    "N2160556120",
                    "N2160556121",
                    "N2160556122",
                    "N2160556123",
                    "N2160556124",
                    "N2160556125",
                    "N2160556126",
                    "N2160556127",
                    "N2160556128",
                ],
                [
                    "N2160556150",
                    "N2160556151",
                    "N2160556152",
                    "N2160556153",
                    "N2160556154",
                    "N2160556155",
                    "N2160556156",
                    "N2160556157",
                    "N2160556158",
                ],
            ]
        )
        result = rhp_geodataframe_with_values.rhp.polyfill(10, explode=True)
        assert len(result) == len(rhp_geodataframe_with_values) * 9
        assert set(result[COLUMNS["polyfill"]]) == expected_indices
        assert not result["val"].isna().any()

    def test_polyfill_explode_unequal_lengths(self, basic_geodataframe_polygons):
        expected_indices = {
            "N2085",
            "N2160",
        }
        result = basic_geodataframe_polygons.rhp.polyfill(4, explode=True)
        assert len(result) == 3
        assert set(result[COLUMNS["polyfill"]]) == expected_indices


class TestCellArea:
    def test_cell_area(self, indexed_dataframe):
        expected = indexed_dataframe.assign(
            rhp_cell_area=[0.258507625363534, 0.258507625363534]
        )
        result = indexed_dataframe.rhp.cell_area()

        pd.testing.assert_frame_equal(expected, result)


class TestLinetrace:
    pass


# Tests: Aggregate functions
class TestGeoToRhpAggregate:
    def test_geo_to_rhp_aggregate(self, basic_dataframe_with_values):
        result = basic_dataframe_with_values.rhp.geo_to_rhp_aggregate(
            1, return_geometry=False
        )
        expected = pd.DataFrame(
            {f"{COLUMNS['prefix']}01": ["N2"], "val": [2 + 5]}
        ).set_index(f"{COLUMNS['prefix']}01")

        pd.testing.assert_frame_equal(expected, result)

    def test_geo_to_rhp_aggregate_geo(self, basic_geodataframe_with_values):
        result = basic_geodataframe_with_values.rhp.geo_to_rhp_aggregate(
            1, return_geometry=False
        )
        expected = pd.DataFrame(
            {f"{COLUMNS['prefix']}01": ["N2"], "val": [2 + 5]}
        ).set_index(f"{COLUMNS['prefix']}01")

        pd.testing.assert_frame_equal(expected, result)

    def test_geo_to_rhp_aggregate_with_geometry(self, basic_dataframe_with_values):
        result = basic_dataframe_with_values.rhp.geo_to_rhp_aggregate(1)
        indexed = pd.DataFrame(
            {f"{COLUMNS['prefix']}01": ["N2"], "val": [2 + 5]}
        ).set_index(f"{COLUMNS['prefix']}01")
        geometry = [
            Polygon(rhp_py.rhp_to_geo_boundary(h, True, False)) for h in indexed.index
        ]
        expected = gpd.GeoDataFrame(indexed, geometry=geometry, crs="epsg:4326")

        assert_geodataframe_equal(expected, result)


class TestRhpToParentAggregate:
    def test_rhp_to_parent_aggregate(self, rhp_geodataframe_with_values):
        result = rhp_geodataframe_with_values.rhp.rhp_to_parent_aggregate(8)

        index = pd.Index(["N21605561"], name=f"{COLUMNS['prefix']}08")
        geometry = [Polygon(rhp_py.rhp_to_geo_boundary(h, True, False)) for h in index]
        expected = gpd.GeoDataFrame(
            {"val": [1 + 2 + 5]}, geometry=geometry, index=index, crs="epsg:4326"
        )

        assert_geodataframe_equal(expected, result)

    def test_rhp_to_parent_aggregate_no_geometry(self, rhp_dataframe_with_values):
        result = rhp_dataframe_with_values.rhp.rhp_to_parent_aggregate(
            8, return_geometry=False
        )
        index = pd.Index(["N21605561"], name=f"{COLUMNS['prefix']}08")
        expected = pd.DataFrame({"val": [1 + 2 + 5]}, index=index)

        pd.testing.assert_frame_equal(expected, result)


class TestPolyfillResample:
    def test_polyfill_resample(self, rhp_geodataframe_with_values):
        expected_indices = set().union(
            *[
                [
                    "N2160556110",
                    "N2160556111",
                    "N2160556112",
                    "N2160556113",
                    "N2160556114",
                    "N2160556115",
                    "N2160556116",
                    "N2160556117",
                    "N2160556118",
                ],
                [
                    "N2160556120",
                    "N2160556121",
                    "N2160556122",
                    "N2160556123",
                    "N2160556124",
                    "N2160556125",
                    "N2160556126",
                    "N2160556127",
                    "N2160556128",
                ],
                [
                    "N2160556150",
                    "N2160556151",
                    "N2160556152",
                    "N2160556153",
                    "N2160556154",
                    "N2160556155",
                    "N2160556156",
                    "N2160556157",
                    "N2160556158",
                ],
            ]
        )
        expected_values = set([1, 2, 5])
        result = rhp_geodataframe_with_values.rhp.polyfill_resample(
            10, return_geometry=False
        )
        assert len(result) == len(rhp_geodataframe_with_values) * 9
        assert set(result.index) == expected_indices
        assert set(result["val"]) == expected_values
        assert not result["val"].isna().any()

    def test_polyfill_resample_uncovered_rows(self, basic_geodataframe_polygons):
        basic_geodataframe_polygons.iloc[1] = box(14, 50, 15, 53)
        with pytest.warns(UserWarning):
            result = basic_geodataframe_polygons.rhp.polyfill_resample(2)

        assert len(result) == 0



from setuptools import setup, find_packages
from os import path

__version__: str = "0.1.0"

# read the contents of README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="rhppandas",
    version=__version__,
    description="Integration of rHEALPixDGGS and Pandas/GeoPandas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="MWLR",
    url="https://github.com/manaakiwhenua/rHP-Pandas",
    keywords=[
        "python",
        "rHEALPix",
        "rhealpixdggs",
        "geospatial",
        "geopandas",
        "pandas",
        "integration",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    packages=find_packages(),
    setup_requires=[],
    install_requires=[
        "geopandas",
        "numpy",
        "pandas",
        "shapely",
        "rhealpixdggs",
        "numpy",
        "typing-extensions",
    ],
    python_requires=">=3.11",
    extras_require={
        "test": ["pytest", "pytest-cov"],
        #     "docs": ["sphinx", "numpydoc", "pytest-sphinx-theme", "typing-extensions"],
    },
)
